@@ -1,15 +1,13 @@
 # flake8: noqa
-__submodules__ = ["construct_bids", "filtering", "input_generation"]
+__submodules__ = ["construct_bids", "filtering", "input_generation", "datasets"]
 
 __ignore__ = ["T_co"]
 
 # <AUTOGEN_INIT>
 from snakebids.core.construct_bids import bids, print_boilerplate
+from snakebids.core.datasets import BidsComponent, BidsDataset, BidsDatasetDict
 from snakebids.core.filtering import filter_list, get_filtered_ziplist_index
 from snakebids.core.input_generation import (
-    BidsComponent,
-    BidsDataset,
-    BidsDatasetDict,
     generate_inputs,
     get_wildcard_constraints,
     write_derivative_json,
