@@ -1,354 +1,32 @@
 """Utilities for converting Snakemake apps to BIDS apps."""
 
-import itertools as it
 import json
 import logging
-import operator as op
 import re
-from collections import UserDict, defaultdict
+from collections import defaultdict
 from pathlib import Path
-from string import Formatter
-from typing import (
-    TYPE_CHECKING,
-    Any,
-    Dict,
-    Generator,
-    Iterable,
-    List,
-    Optional,
-    Tuple,
-    Union,
-    cast,
-    overload,
-)
+from typing import Dict, Generator, Iterable, List, Optional, Tuple, Union, overload
 
-import attr
 import more_itertools as itx
 from bids import BIDSLayout, BIDSLayoutIndexer
 from bids.layout import Query
-from cached_property import cached_property
-from typing_extensions import Literal, TypedDict
+from typing_extensions import Literal
 
-import snakebids.utils.sb_itertools as sb_it
+from snakebids.core.datasets import BidsComponent, BidsDataset, BidsDatasetDict
 from snakebids.core.filtering import filter_list
 from snakebids.exceptions import ConfigError
+from snakebids.types import InputsConfig
 from snakebids.utils.snakemake_io import glob_wildcards
-from snakebids.utils.utils import BidsEntity, property_alias
+from snakebids.utils.utils import BidsEntity, BidsParseError, surround
 
 _logger = logging.getLogger(__name__)
 
 
-class BidsDatasetDict(TypedDict):
-    """Dict equivalent of BidsInputs, for backwards-compatibility"""
-
-    input_path: Dict[str, str]
-    input_zip_lists: Dict[str, Dict[str, List[str]]]
-    input_lists: Dict[str, Dict[str, List[str]]]
-    input_wildcards: Dict[str, Dict[str, str]]
-    subjects: List[str]
-    sessions: List[str]
-    subj_wildcards: Dict[str, str]
-
-
-@attr.define
-class BidsComponent:
-    """Component of a BidsDataset mapping entities to their resolved values
-
-    BidsComponents are immutable: their values cannot be altered.
-
-    Properties
-    ----------
-    name
-        Name of the component
-    path
-        Wildcard-filled path that matches the files for this component.
-    zip_lists
-        Dictionary where each key is a wildcard entity and each value is a list of the
-        values found for that entity. Each of these lists has length equal to the number
-        of images matched for this modality, so they can be zipped together to get a
-        list of the wildcard values for each file.
-    """
-
-    name: str = attr.field(on_setattr=attr.setters.frozen)
-    path: str = attr.field(on_setattr=attr.setters.frozen)
-    zip_lists: Dict[str, List[str]] = attr.field(
-        on_setattr=attr.setters.frozen, converter=dict
-    )
-
-    @zip_lists.validator  # type: ignore
-    def _validate_zip_lists(self, _, value: Dict[str, List[str]]):
-        lengths = {len(val) for val in value.values()}
-        if len(lengths) > 1:
-            raise ValueError("zip_lists must all be of equal length")
-        _, raw_fields, *_ = sb_it.unpack(
-            zip(*Formatter().parse(self.path)), [[], [], []]
-        )
-        fields = set(filter(None, raw_fields))
-        if fields != set(value):
-            raise ValueError(
-                "zip_lists entries must match the wildcards in input_path: "
-                f"Path: {fields} != zip_lists: {set(value)}"
-            )
-
-    @property
-    def input_name(self):
-        """Alias of :attr:`name <snakebids.BidsComponent.name>`
-
-        Name of the component
-        """
-        return self.name
-
-    @property
-    def input_path(self):
-        """Alias of :attr:`path <snakebids.BidsComponent.path>`
-
-        Wildcard-filled path that matches the files for this component.
-        """
-        return self.path
-
-    @property
-    def input_zip_lists(self):
-        """Alias of :attr:`zip_lists <snakebids.BidsComponent.zip_lists>`
-
-        Dictionary where each key is a wildcard entity and each value is a list of the
-        values found for that entity. Each of these lists has length equal to the number
-        of images matched for this modality, so they can be zipped together to get a
-        list of the wildcard values for each file.
-        """
-        return self.zip_lists
-
-    # Note: we can't use cached property here because it's incompatible with slots.
-    _input_lists: Optional[Dict[str, List[str]]] = attr.field(
-        default=None, init=False, eq=False, repr=False
-    )
-    _input_wildcards: Optional[Dict[str, str]] = attr.field(
-        default=None, init=False, eq=False, repr=False
-    )
-    _entities: Optional[List[str]] = attr.field(
-        default=None, init=False, eq=False, repr=False
-    )
-
-    @property
-    def entities(self):
-        """Component entities and their associated values
-
-        Dictionary where each key is an entity and each value is a list of the
-        unique values found for that entity. These lists might not be the same length.
-        """
-        if self._input_lists is None:
-            self._input_lists = {
-                entity: list(set(values)) for entity, values in self.zip_lists.items()
-            }
-        return self._input_lists
-
-    @property_alias(entities, "entities", "snakebids.BidsComponent.entities")
-    def input_lists(self):
-        return self.entities
-
-    @property
-    def wildcards(self):
-        """Wildcards in brace-wrapped syntax
-
-        Dictionary where each key is the name of a wildcard entity, and each value is
-        the Snakemake wildcard used for that entity.
-        """
-        if self._input_wildcards is None:
-            self._input_wildcards = {
-                entity: f"{{{entity}}}" for entity in self.zip_lists
-            }
-        return self._input_wildcards
-
-    @property_alias(wildcards, "wildcards", "snakebids.BidsComponent.wildcards")
-    def input_wildcards(self):
-        return self.wildcards
-
-    def __eq__(self, other: Union["BidsComponent", object]):
-        if not isinstance(other, BidsComponent):
-            return False
-
-        if self.name != other.name:
-            return False
-
-        if self.path != other.path:
-            return False
-
-        def sorted_items(dictionary: Dict[str, List[str]]):
-            return sorted(dictionary.items(), key=op.itemgetter(0))
-
-        if set(self.zip_lists) != set(other.zip_lists):
-            return False
-
-        if not other.zip_lists and not self.zip_lists:
-            return True
-
-        other_items = cast(
-            List[List[str]], list(zip(*sorted_items(other.zip_lists)))[1]
-        )
-        our_items = cast(List[List[str]], list(zip(*sorted_items(self.zip_lists)))[1])
-
-        return set(zip(*our_items)) == set(zip(*other_items))
-
-
-if TYPE_CHECKING:
-    _BidsComponentsType = UserDict[str, BidsComponent]
-else:
-    # UserDict is not subscriptable in py37
-    _BidsComponentsType = UserDict
-
-
-class BidsDataset(_BidsComponentsType):
-    """A bids dataset parsed by pybids, organized into BidsComponents.
-
-    BidsDatasets are typically generated using :func:`generate_inputs()
-    <snakebids.generate_inputs>`, which reads the ``pybids_inputs`` field in your
-    snakemake config file and, for each entry, creates a BidsComponent using the
-    provided name, wildcards, and filters.
-
-    Individual components can be accessed using bracket-syntax: (e.g.
-    ``inputs["t1w"]``). Component access attributes along with the component name in
-    brackets can also be used. For example, ``BidsDataset.entities["t1w"]`` and
-    ``BidsDataset["t1w"].entities`` return the same thing.
-
-    Provides access to summarizing information, for instance, the set of all subjects or
-    sessions found in the dataset
-    """
-
-    # pylint: disable=super-init-not-called
-    def __init__(self, data: Any):
-        self.data = dict(data)
-
-    def __setitem__(self, _: Any, __: Any):
-        raise NotImplementedError(
-            f"Modification of {self.__class__.__name__} is not yet supported"
-        )
-
-    @cached_property
-    def path(self):
-        """Dict mapping :class:`BidsComponents <snakebids.BidsComponent>` names to \
-        their ``paths``.
-        """
-        return {key: value.path for key, value in self.data.items()}
-
-    @property_alias(path, "path", "snakebids.BidsDataset.path")
-    def input_path(self):
-        return self.path
-
-    @cached_property
-    def zip_lists(self):
-        """Dict mapping :class:`BidsComponents <snakebids.BidsComponent>` names to \
-        their ``zip_lists``
-        """
-        return {key: value.zip_lists for key, value in self.data.items()}
-
-    @property_alias(zip_lists, "zip_lists", "snakebids.BidsDataset.zip_lists")
-    def input_zip_lists(self):
-        return self.zip_lists
-
-    @cached_property
-    def entities(self):
-        """Dict mapping :class:`BidsComponents <snakebids.BidsComponent>` names to \
-        to their :attr:`entities <snakebids.BidsComponent.entities>`
-        """
-        return {key: value.entities for key, value in self.data.items()}
-
-    @property_alias(entities, "entities", "snakebids.BidsDataset.entities")
-    def input_lists(self):
-        return self.entities
-
-    @cached_property
-    def wildcards(self):
-        """Dict mapping :class:`BidsComponents <snakebids.BidsComponent>` names to \
-        their :attr:`wildcards <snakebids.BidsComponent.wildcards>`
-        """
-        return {key: value.input_wildcards for key, value in self.data.items()}
-
-    @property_alias(wildcards, "wildcards", "snakebids.BidsDataset.wildcards")
-    def input_wildcards(self):
-        return self.wildcards
-
-    @cached_property
-    def subjects(self):
-        """A list of the subjects in the dataset."""
-        return [
-            *{
-                *it.chain.from_iterable(
-                    input_list["subject"]
-                    for input_list in self.entities.values()
-                    if "subject" in input_list
-                )
-            }
-        ]
-
-    @property
-    # @ft.lru_cache(None)
-    def sessions(self):
-        """A list of the sessions in the dataset."""
-        return [
-            *{
-                *it.chain.from_iterable(
-                    input_list["session"]
-                    for input_list in self.entities.values()
-                    if "session" in input_list
-                )
-            }
-        ]
-
-    @property
-    # @ft.lru_cache(None)
-    def subj_wildcards(self):
-        """The subject and session wildcards applicable to this dataset.
-
-        ``{"subject":"{subject}"}`` if there is only one session, ``{"subject":
-        "{subject}", "session": "{session}"}`` if there are multiple sessions.
-        """
-        if len(self.sessions) == 0:
-            return {"subject": "{subject}"}
-        return {
-            "subject": "{subject}",
-            "session": "{session}",
-        }
-
-    @property
-    def as_dict(self):
-        """Get the layout as a legacy dict
-
-        Included primarily for backward compatability with older versions of snakebids,
-        where generate_inputs() returned a dict rather than the `BidsDataset` class
-
-        Returns
-        -------
-        BidsDatasetDict
-        """
-        return BidsDatasetDict(
-            input_path=self.input_path,
-            input_lists=self.entities,
-            input_wildcards=self.input_wildcards,
-            input_zip_lists=self.input_zip_lists,
-            subjects=self.subjects,
-            sessions=self.sessions,
-            subj_wildcards=self.subj_wildcards,
-        )
-
-    @classmethod
-    def from_iterable(cls, iterable: Iterable[BidsComponent]):
-        """Construct Dataset from iterable of BidsComponents
-
-        Parameters
-        ----------
-        iterable : Iterable[BidsComponent]
-
-        Returns
-        -------
-        BidsDataset
-        """
-        return cls({bidsinput.name: bidsinput for bidsinput in iterable})
-
-
 # pylint: disable=too-many-arguments
 @overload
 def generate_inputs(
     bids_dir,
-    pybids_inputs,
+    pybids_inputs: InputsConfig,
     pybids_database_dir=...,
     pybids_reset_database=...,
     derivatives=...,
@@ -365,7 +43,7 @@ def generate_inputs(
 @overload
 def generate_inputs(
     bids_dir,
-    pybids_inputs,
+    pybids_inputs: InputsConfig,
     pybids_database_dir=...,
     pybids_reset_database=...,
     derivatives=...,
@@ -378,10 +56,10 @@ def generate_inputs(
     ...
 
 
-# pylint: disable=too-many-arguments
+# pylint: disable=too-many-arguments, too-many-locals
 def generate_inputs(
     bids_dir,
-    pybids_inputs,
+    pybids_inputs: InputsConfig,
     pybids_database_dir=None,
     pybids_reset_database=False,
     derivatives=False,
@@ -452,16 +130,17 @@ def generate_inputs(
         be specified if participant_label is specified
 
     use_bids_inputs : bool, optional
-        If True, opts in to the new BidsInputs output, otherwise returns the classic
-        dict. Currently, the classic dict will be returned by default, however, this
-        will change in a future release. If you do not wish to migrate to the new
-        BidsInputs, we recommend you explictely set this parameter to False
+        If True, opts in to the new :class:`BidsDataset` output, otherwise returns the
+        classic dict. Currently, the classic dict will be returned by default, however,
+        this will change in a future release. If you do not wish to migrate to the new
+        BidsDataset, we recommend you explictely set this parameter to False
 
     Returns
     -------
     BidsDataset or BidsDatasetDict:
-        Object containing organized information about the bids inputs for consumption
-        in snakemake. See the documentation of BidsInputs for details and examples.
+        Object containing organized information about the bids inputs for consumption in
+        snakemake. See the documentation of :class:`BidsDataset` for details and
+        examples.
 
     Example
     -------
@@ -544,9 +223,9 @@ def generate_inputs(
             "subj_wildcards": {"subject": "{subject}"}
         }
 
-    Or, if BidsInputs is enabled::
+    Or, if :class:`BidsDataset` is enabled::
 
-        <class BidsInputs>
+        <class BidsDataset>
             path: {
                 "bold": "bids-example/sub-{subject}/func/sub-{subject}_task-{task}_bold\
                     .nii.gz"
@@ -579,12 +258,16 @@ def generate_inputs(
     )
 
     # Generates a BIDSLayout
-    layout = _gen_bids_layout(
-        bids_dir=bids_dir,
-        derivatives=derivatives,
-        pybids_config=pybids_config,
-        pybids_database_dir=pybids_database_dir,
-        pybids_reset_database=pybids_reset_database,
+    layout = (
+        _gen_bids_layout(
+            bids_dir=bids_dir,
+            derivatives=derivatives,
+            pybids_config=pybids_config,
+            pybids_database_dir=pybids_database_dir,
+            pybids_reset_database=pybids_reset_database,
+        )
+        if not _all_custom_paths(pybids_inputs)
+        else None
     )
 
     filters = {"subject": subject_filter} if subject_filter else {}
@@ -599,28 +282,42 @@ def generate_inputs(
     if use_bids_inputs is None:
         _logger.warning(
             "The dictionary returned by generate_inputs() will soon be deprecated in "
-            "favour of the new BidsInputs class. BidsInputs provides the same "
+            "favour of the new BidsDataset class. BidsDataset provides the same "
             "functionality of the dict, but with attribute access and new convience "
             "methods. In a future release, generate_inputs() will return this class "
             "by default. You can opt into this behaviour now by setting the "
             "`use_bids_inputs` argument in generate_inputs() to True. If you do not "
-            "wish to migrate your code to BidsInputs at this time, we recommend you "
+            "wish to migrate your code to BidsDataset at this time, we recommend you "
             "explicately set `use_bids_inputs` to False. This will preserve the "
             "current behaviour of returning a Dict in future releases, and will "
             "silence this warning."
         )
         use_bids_inputs = False
+
+    try:
+        dataset = BidsDataset.from_iterable(bids_inputs)
+    except ValueError as err:
+        raise ConfigError(
+            "Multiple components found with the same name: "
+            + ", ".join(surround(err.args[0], "'"))
+        ) from err
+
     if use_bids_inputs:
-        return BidsDataset.from_iterable(bids_inputs)
-    return BidsDataset.from_iterable(bids_inputs).as_dict
+        return dataset
+    return dataset.as_dict
+
+
+def _all_custom_paths(config: InputsConfig):
+    """Check that all input components have a custom path"""
+    return all(comp.get("custom_path") for comp in config.values())
 
 
 def _gen_bids_layout(
-    bids_dir,
-    derivatives,
-    pybids_database_dir,
-    pybids_reset_database,
-    pybids_config=None,
+    bids_dir: Union[Path, str],
+    derivatives: bool,
+    pybids_database_dir: Union[Path, str, None],
+    pybids_reset_database: bool,
+    pybids_config: Union[Path, str, None] = None,
 ):
     """Create (or reindex) the BIDSLayout if one doesn't exist,
     which is only saved if a database directory path is provided
@@ -649,37 +346,26 @@ def _gen_bids_layout(
         Layout from pybids for accessing the BIDS dataset to grab paths
     """
 
-    # Set db dir to None (otherwise saves to parent dir)
-    if Path(bids_dir):
-        # Check for database_dir
-        # If blank, assume db not to be used
-        if pybids_database_dir == "":
-            pybids_database_dir = None
-        # Otherwise check for relative path and update
-        elif (
-            pybids_database_dir is not None
-            and not Path(pybids_database_dir).is_absolute()
-        ):
-            pybids_database_dir = None
-            _logger.warning("Absolute path must be provided, database will not be used")
+    # Check for database_dir
+    # If blank, assume db not to be used
+    if pybids_database_dir == "":
+        pybids_database_dir = None
+    # Otherwise check for relative path and update
+    elif (
+        pybids_database_dir is not None and not Path(pybids_database_dir).is_absolute()
+    ):
+        pybids_database_dir = None
+        _logger.warning("Absolute path must be provided, database will not be used")
 
-        layout = BIDSLayout(
-            bids_dir,
-            derivatives=derivatives,
-            validate=False,
-            config=pybids_config,
-            database_path=pybids_database_dir,
-            reset_database=pybids_reset_database,
-            indexer=BIDSLayoutIndexer(validate=False, index_metadata=False),
-        )
-    else:
-        _logger.info(
-            "bids_dir does not exist, skipping PyBIDS and using "
-            "custom file paths only"
-        )
-        layout = None
-
-    return layout
+    return BIDSLayout(
+        bids_dir,
+        derivatives=derivatives,
+        validate=False,
+        config=pybids_config,
+        database_path=pybids_database_dir,
+        reset_database=pybids_reset_database,
+        indexer=BIDSLayoutIndexer(validate=False, index_metadata=False),
+    )
 
 
 def write_derivative_json(snakemake, **kwargs):
@@ -839,50 +525,28 @@ def _parse_bids_path(path: str, entities: Iterable[str]) -> Tuple[str, Dict[str,
 
     wildcard_values: Dict[str, str] = {}
 
-    # BUG: No support for "datatype"
     for entity in map(BidsEntity, entities):
         # Iterate over wildcards, slowly updating the path as each entity is replaced
 
-        tag = entity.tag
         wildcard = entity.wildcard
+        match = re.search(entity.regex, path)
+        if not match or not match.group(2):
+            raise BidsParseError(path=path, entity=entity)
 
-        if entity == "suffix":
-            # capture suffix
-            match = re.search(r".*_([a-zA-Z0-9]+).*$", path)
-
-            # capture "(before)suffix(after)" and replace with "before{suffix}after"
-            new_path = re.sub(r"(.*_)[a-zA-Z0-9]+(.*)$", rf"\1{{{wildcard}}}\2", path)
-
-        else:
-            # BUG: tags cannot be a substr of another tag in the same path (e.g. space
-            #      and ce)
-            pattern = f"{tag}-([a-zA-Z0-9]+)"
-            replace = f"{tag}-{{{wildcard}}}"
-
-            match = re.search(pattern, path)
-            new_path = re.sub(pattern, replace, path)
-
-        if match and match.group(1):
-            value = match.group(1)
-        else:
-            value = ""
-
-        # update the path with the {wildcards} -- uses the
-        # value from the string (not from the pybids
-        # entities), since that has issues with integer
-        # formatting (e.g. for run=01)
-
-        path = new_path
+        # overwrite path one wildcard at a time
+        path = re.sub(entity.regex, rf"\1{{{wildcard}}}\3", path)
+        value = match.group(2)
         wildcard_values[wildcard] = value
 
     return path, wildcard_values
 
 
+# pylint: disable=too-many-locals
 def _get_lists_from_bids(
     bids_layout: Optional[BIDSLayout],
-    pybids_inputs,
+    pybids_inputs: InputsConfig,
     *,
-    limit_to=None,
+    limit_to: "list[str] | None" = None,
     **filters,
 ) -> Generator[BidsComponent, None, None]:
     """Grabs files using pybids and creates snakemake-friendly lists
@@ -905,25 +569,26 @@ def _get_lists_from_bids(
 
     Yields
     ------
-    BidsLists:
-        One BidsLists is yielded for each modality described by ``pybids_inputs``.
+    BidsComponent:
+        One BidsComponent is yielded for each modality described by ``pybids_inputs``.
     """
     if limit_to is None:
-        limit_to = pybids_inputs.keys()
+        limit_to = list(pybids_inputs)
 
     for input_name in limit_to:
         _logger.debug("Grabbing inputs for %s...", input_name)
+        component = pybids_inputs[input_name]
 
-        if "custom_path" in pybids_inputs[input_name].keys():
+        if "custom_path" in component:
             # a custom path was specified for this input, skip pybids:
             # get input_wildcards by parsing path for {} entries (using a set
             # to get unique only)
             # get zip_lists by using glob_wildcards (but need to modify
             # to deal with multiple wildcards
 
-            path: str = pybids_inputs[input_name]["custom_path"]
+            path = component["custom_path"]
             zip_lists = _parse_custom_path(
-                path, **pybids_inputs[input_name]["filters"], **filters
+                path, **pybids_inputs[input_name].get("filters", {}), **filters
             )
             yield BidsComponent(input_name, path, zip_lists)
             continue
@@ -940,17 +605,36 @@ def _get_lists_from_bids(
         paths = set()
         pybids_filters = {
             key: Query.ANY if val is True else Query.NONE if val is False else val
-            for key, val in pybids_inputs[input_name].get("filters", {}).items()
+            for key, val in component.get("filters", {}).items()
         }
         for img in bids_layout.get(**pybids_filters, **filters):
             wildcards: List[str] = [
                 wildcard
-                for wildcard in pybids_inputs[input_name].get("wildcards", [])
-                if wildcard in img.get_entities()
+                for wildcard in component.get("wildcards", [])
+                if wildcard in img.entities
             ]
             _logger.debug("Wildcards %s found entities for %s", wildcards, img.path)
 
-            path, parsed_wildcards = _parse_bids_path(img.path, wildcards)
+            try:
+                path, parsed_wildcards = _parse_bids_path(img.path, wildcards)
+            except BidsParseError as err:
+                raise ConfigError(
+                    "Parsing failed:\n"
+                    f"  Entity: {err.entity.entity}\n"
+                    f"  Pattern: {err.entity.regex}\n"
+                    f"  Path: {img.path}\n"
+                    "\n"
+                    "Pybids parsed this path using the pattern: "
+                    f"{bids_layout.entities[err.entity.entity].regex}\n"
+                    "\n"
+                    "Snakebids is not currently able to handle this entity. If it is a "
+                    "custom entity, its `tag-` must be configured to be the same as "
+                    "its name. Its entry in your pybids config file should look like:\n"
+                    f'{{\n\t"name": "{err.entity.entity}",\n'
+                    f'\t"pattern":"{err.entity.entity}-(<value_pattern>)"\n}}\n'
+                    f"If {err.entity.entity} is an official pybids entity, please "
+                    "ensure you are using the latest version of snakebids"
+                ) from err
 
             for wildcard_name, value in parsed_wildcards.items():
                 zip_lists[wildcard_name].append(value)
@@ -959,7 +643,7 @@ def _get_lists_from_bids(
 
         # now, check to see if unique
         if len(paths) == 0:
-            _logger.warning("No images found for %s", input_name)
+            _logger.warning("No input file found for %s", input_name)
             continue
         try:
             path = itx.one(paths)
