@@ -3,9 +3,11 @@ from __future__ import annotations
 import functools as ft
 import importlib.resources
 import json
+import re
 from typing import Any, Callable, Dict, Iterable, TypeVar
 
 import attrs
+import more_itertools as itx
 from typing_extensions import Protocol
 
 from snakebids.utils.sb_typing import UserProperty
@@ -76,8 +78,8 @@ class BidsEntity:
         )
 
     @property
-    def pattern(self):
-        """Get regex of acceptable patterns
+    def match(self):
+        """Get regex of acceptable value matches
 
         If no pattern is associated with the entity, the default pattern is a word with
         letters and numbers
@@ -88,11 +90,53 @@ class BidsEntity:
         """
         tags = read_bids_tags()
         return (
-            tags[self.entity]["pattern"]
-            if self.entity in tags and "pattern" in tags[self.entity]
+            tags[self.entity]["match"]
+            if self.entity in tags and "match" in tags[self.entity]
             else "[a-zA-Z0-9]+"
         )
 
+    @property
+    def before(self):
+        """regex str to search before value in paths
+
+        Returns
+        -------
+        str
+        """
+        tags = read_bids_tags()
+        return (
+            tags[self.entity]["before"]
+            if self.entity in tags and "before" in tags[self.entity]
+            else f"{self.tag}-"
+        )
+
+    @property
+    def after(self):
+        """regex str to search after value in paths
+
+        Returns
+        -------
+        str
+        """
+        tags = read_bids_tags()
+        return (
+            tags[self.entity]["after"]
+            if self.entity in tags and "after" in tags[self.entity]
+            else ""
+        )
+
+    @property
+    def regex(self):
+        """Complete pattern to match when searching in paths
+
+        Contains three capture groups, the first corresponding to "before", the second
+        to "value", and the third to "after"
+        Returns
+        -------
+        str
+        """
+        return re.compile(f"({self.before})({self.match})({self.after})")
+
     @property
     def wildcard(self):
         """Get the snakebids {wildcard}
@@ -174,6 +218,15 @@ def get_match_search_func(
     return inner
 
 
+class BidsParseError(Exception):
+    """Exception raised for errors encountered in the parsing of Bids paths"""
+
+    def __init__(self, path: str, entity: BidsEntity):
+        self.path = path
+        self.entity = entity
+        super().__init__(path, entity)
+
+
 # pylint: disable=too-few-public-methods
 class _Documented(Protocol):
     __doc__: str
@@ -205,3 +258,9 @@ def property_alias(prop: _Documented, label: str | None = None, ref: str | None
         return alias
 
     return inner
+
+
+def surround(__s: Iterable[str] | str, __object: str) -> Iterable[str]:
+    """Surround a string or each string in an iterable with characters"""
+    for item in itx.always_iterable(__s):
+        yield __object + item + __object
